clear all; close all;

handLoadControls = true;
springPosParams = false;
% addpath('C:\Files\GPOPS-II New\nlp\snopt')

% Load model parameters determined by past calibration runs
cdorig = cd;
cd('../SetupFiles')
auxdata = SetupOpenSimOptimalTracking(1,true);
cd(cdorig)

auxdata.handLoadControls = handLoadControls;
auxdata.springPosParams = springPosParams;

% Load initial and final model pose and time
t0 = auxdata.time(1);
tf = auxdata.time(end);
time = auxdata.time;
q0 = auxdata.q(1,:);
qp0 = auxdata.qp(1,:);
qf = auxdata.q(end,:);
qpf = auxdata.qp(end,:);

% Set initial guesses for Loads (control) and joint angle trajectories
% (states)
loadsGuess = auxdata.idData;
qGuess = auxdata.q;
qpGuess = auxdata.qp;
qppGuess = auxdata.qpp;
qpppGuess = auxdata.qppp;
nJerkControls = size(qpppGuess,2);
auxdata.nJerkControls = nJerkControls;

% Determine the range of the states
qRange = max(qGuess) - min(qGuess);
qpRange = max(qpGuess) - min(qpGuess);
qppRange = max(qppGuess) - min(qppGuess);
qpppRange = max(qpppGuess) - min(qpppGuess);

auxdata.activeDofs = logical(1-(qRange == 0));
auxdata.activeDofs(1:6) = [];

% Give locked DOFs a range of motion
qpppRange(qRange == 0) = 40000*pi/180;
qppRange(qRange == 0) = 4000*pi/180;
qpRange(qRange == 0) = 400*pi/180;
qRange(qRange == 0) = 60*pi/180;

% Set bounds on the states based on the previously calcualted range
qMin = min(qGuess) - 2*qRange;  
qMax  = max(qGuess) + 2*qRange;
qpMin = min(qpGuess) - 1.5*qpRange; 
qpMax  = max(qpGuess) + 1.5*qpRange;
qppMin = min(qppGuess) - 1*qppRange; 
qppMax  = max(qppGuess) + 1*qppRange;
qpppMin = min(qpppGuess) - 1*qpppRange; % 0.5 %NOTE (Nathan): This is different than previous step (ground contact calibration)
qpppMax  = max(qpppGuess) + 1*qpppRange; %0.5

% Determing the range of the controls
loadsRange = max(auxdata.idData) - min(auxdata.idData);
torquesRange = max(auxdata.torques) - min(auxdata.torques);

% Give bigger range to DOFs with small loads (for instance the toes)
loadsRange(loadsRange<1/(auxdata.tScale^2)) = 25/(auxdata.tScale^2);
torquesRange(torquesRange<1/(auxdata.tScale^2)) = 25/(auxdata.tScale^2);

% Set bounds on the torques (for when using extra controls)
loadsMin = min(auxdata.idData)-2*loadsRange; 
loadsMax = max(auxdata.idData)+2*loadsRange;

torquesMin = min(auxdata.torques) - 2*torquesRange;
torquesMax = max(auxdata.torques) + 2*torquesRange;

torquesGuess = auxdata.torques;

% Set weights for marker errors (the model tracks markers from experimental
% data
% MkrNamesOSIM = {'R.Shoulder' 'L.Shoulder' 'Neck' 'Chest' 'R.Elbow' 'R.Wrist' 'L.Elbow' 'L.Wrist' 'R.ASIS' 'L.ASIS'...
%     'Sacral' 'R.Thigh.Superior' 'R.Thigh.Inferior' 'R.Thigh.Lateral' 'R.Shank.Superior' 'R.Shank.Inferior' 'R.Shank.Lateral'...
%     'R.Heel' 'R.Midfoot.Superior' 'R.Midfoot.Lateral' 'R.Toe' 'L.Thigh.Superior' 'L.Thigh.Inferior' 'L.Thigh.Lateral'...
%     'L.Shank.Superior' 'L.Shank.Lateral' 'L.Shank.Inferior' 'L.Heel' 'L.Midfoot.Superior' 'L.Midfoot.Lateral' 'L.Toe'...
%     'L.PSIS' 'R.PSIS'};
auxdata.errorWeights = ones(1,numel(auxdata.mkrNames));
auxdata.errorWeights([18 21 28 31]) = 10;
auxdata.errorWeights([3:8]) = .1;

auxdata.errorWeights = repmat(auxdata.errorWeights,3,1);
auxdata.errorWeights = auxdata.errorWeights(:)';

% Initilize osim models in mex files
opensimPointKin_v3('Subject1_v4pt3_GPOPS.osim');
opensimID_OMP('Subject1_v4pt3_GPOPS.osim');
opensimGetBodyPosVel_OMP('Subject1_v4pt3_GPOPS.osim');

% Get median hand positions
handForceLoc = [0.01,-0.06,0; 0.01,-0.06,0]; % NOTE (Nathan): Make sure that this is the same location as where the force is applied in the model
auxdata.handForceLoc = handForceLoc;
handBodyMat = auxdata.handBodyMat;

auxdata.posOnBarR = [0.874274080385552,1.12880647759450,0.602273837458135]; %NOTE (Nathan): These come from the getWristMarkerLocs program in SetupFiles
auxdata.posOnBarL = [0.869472149835813,1.13141862779562,0.164197648932865];

% [handPos, handVel, handAng, handAngVel] = opensimGetBodyPosVel_OMP(time,[qGuess qpGuess],handForceLoc,handBodyMat);


% Min and Max hand loads allowed
% NOTE (Nathan): Thses are only used if the hand loads are minimized or for
% hand load controls
handLoadMin = -[200*ones(1,3),50*ones(1,3),200*ones(1,3),50*ones(1,3)];
handLoadMax = [200*ones(1,3),50*ones(1,3),200*ones(1,3),50*ones(1,3)];
handLoadRange = handLoadMax - handLoadMin;


% % Marker Error
% % NOTE (Nathan): I don't think these do anything
% mkrErrorsAllowed = 2e-2*ones(1,numel(auxdata.mkrNames)); %Andrew had 1e-2
% mkrErrorsAllowed([18:21 28:31]) = 1.5e-2; %Andrew had 2.5e-3
% mkrErrorsAllowed([3:8]) = 3e-2; %Andrew had 1.5e-2
% mkrErrorsAllowed = repmat(mkrErrorsAllowed, 3,1);% Duplicate for each dimension
% mkrErrorsAllowed = mkrErrorsAllowed(:)';

% GRF Errors
grfErrorsInt = 1e2*[1e1 2e1 1e1 1e1 2e1 1e1 5*ones(1,3) 5*ones(1,3)]; %5e-1* ... 5 %NOTE (Nathan): The torques are different than previous steps %Used for setting path constraint on max error of a GRF
grfErrorsMax = 3*grfErrorsInt; % 2 works better than 3 some times %Used for setting path constraint on max error of a GRF

% Coordinate Errors
qErrorsInt = 5e1*(pi/180)*ones(1,31);
% qErrorsInt(1:3) = 1*qRange(1:3); %1*
qErrorsInt(4:6) = 0.1; %100
% qErrorsInt([9 13 16 20:31]) = 0.25*qRange([9 13 16 20:31]); %0.25*
% qErrorsInt([9 13 16 20]) = 0.25*qRange([9 13 16 20]); %0.25*
% qErrorsInt(21:31) = 0.5*qRange(21:31);


qErrorsMax = 5*qErrorsInt; %5
qErrorsMax([13,20]) = (10*pi/180)*ones(1,2);

% Belt Speed and Number of Springs
auxdata.beltSpeed = 0.6;
auxdata.numSprings = sum(auxdata.numSpringsBody);


% Setup scale factors
auxdata.tMin = t0;
auxdata.tMax = tf;
auxdata.stateMin = [qMin,qpMin,qppMin];
auxdata.stateMax = [qMax,qpMax,qppMax];
if handLoadControls
    auxdata.controlMin = [qpppMin,zeros(1,10),handLoadMin];
    auxdata.controlMax = [qpppMax,1*ones(1,10),handLoadMax];
    auxdata.integralMin = 0*ones(1,31+6+6+12+10+12+31);
    auxdata.integralMax = auxdata.tMax*([qErrorsInt ...
        grfErrorsInt...
        1*loadsRange([7:10,12:13,15:18,20:21])... % [7:10,12:13,15:18,20:21,23:29,33:36] Ignore pelvis, toes, wrist, patella loads. OLD: [1:6,8:13,15:end]
        0.5*ones(1,10)... % For minimizing synergy commands (in case of indeterminicy or something)
        handLoadRange...
        2*qpppRange]).^2; % NOTE (Nathan): This was 1* in previous step % Ignore locked coordinates (wrists) and patella OLD: all
    
else
    auxdata.controlMin = [qpppMin,zeros(1,10)];
    auxdata.controlMax = [qpppMax,1*ones(1,10)];
    auxdata.integralMin = 0*ones(1,31+6+6+12+72+12+31);
    auxdata.integralMax = auxdata.tMax*([qErrorsInt ...
        grfErrorsInt...
        1*loadsRange([7:10,12:13,15:18,20:21])... % [7:10,12:13,15:18,20:21,23:29,33:36] Ignore pelvis, toes, wrist, patella loads. OLD: [1:6,8:13,15:end]
        0.5*(0.25+max(auxdata.acts))...
        handLoadRange ...
        1*qpppRange]).^2; % NOTE (Nathan): This was 1* in previous step % Ignore locked coordinates (wrists) and patella OLD: all
end

onescol = ones(length(auxdata.time),1);
numTorques = size(auxdata.torques,2);
pathMult = 1e0;
auxdata.pathMin = [pathMult*[-1e-1*ones(1,3) -1e0*ones(1,3) -1e-1*ones(1,10)] -qErrorsMax -grfErrorsMax -0.005*ones(1,6)];
auxdata.pathMax = [pathMult*[1e-1*ones(1,3) 1e0*ones(1,3) 1e-1*ones(1,10)] qErrorsMax grfErrorsMax 0.005*ones(1,6)];

periodicityMult = 1e0;
if handLoadControls
    auxdata.eventMin = [periodicityMult*[zeros(1,31)-0.5*pi/180 zeros(1,31)-1*pi/180 -2*[10 25 10 10 25 10]] pathMult*[-1e-1*ones(1,3) -1e0*ones(1,3)]]; % NOTE (Nathan): The 2* wasn't there last step
    auxdata.eventMax = [periodicityMult*[zeros(1,31)+0.5*pi/180 zeros(1,31)+1*pi/180 2*[10 25 10 10 25 10]] pathMult*[1e-1*ones(1,3) 1e0*ones(1,3)]]; % NOTE (Nathan): The 2* wasn't there last step
else
    auxdata.eventMin = [zeros(1,25)-0.5*pi/180 zeros(1,25)-1*pi/180 -2*[10 25 10 10 25 10], -2*[10 25 10 10 25 10] pathMult*[-1e-1*ones(1,3) -1e0*ones(1,3)]]; % NOTE (Nathan): The 2* wasn't there last step
    auxdata.eventMax = [zeros(1,25)+0.5*pi/180 zeros(1,25)+1*pi/180 2*[10 25 10 10 25 10], 2*[10 25 10 10 25 10] pathMult*[1e-1*ones(1,3) 1e0*ones(1,3)]]; % NOTE (Nathan): The 2* wasn't there last step
end



%% Using output from tracking problem
% input.phase.time    = ([auxdata.time]-(auxdata.tMax+auxdata.tMin)/2)./(auxdata.tMax-auxdata.tMin);
% input.phase.state   = ([qGuess qpGuess qppGuess]-onescol*(auxdata.stateMax+auxdata.stateMin)/2)./(onescol*(auxdata.stateMax-auxdata.stateMin));% auxdata.Loads];% auxdata.Loadsp]
% input.phase.control = ([qpppGuess,auxdata.synergyCommands_r,auxdata.synergyCommands_l,auxdata.handLoadsR,auxdata.handLoadsL]-onescol*(auxdata.controlMax+auxdata.controlMin)/2)./(onescol*(auxdata.controlMax-auxdata.controlMin));
% input.auxdata = auxdata;
load('output_synergyCalibration_snopt_c1v4pt5.mat')
solution = output.result.solution;
auxOut = output.result.setup.auxdata;
onesCol = ones(length(solution.phase.time),1);
time_unscaled = (solution.phase.time)*(auxOut.tMax-auxOut.tMin)+(auxOut.tMax+auxOut.tMin)/2;
state_unscaled = (solution.phase.state).*(onesCol*(auxOut.stateMax-auxOut.stateMin))+onesCol*(auxOut.stateMax+auxOut.stateMin)/2;
control_unscaled = (solution.phase.control).*(onesCol*(auxOut.controlMax-auxOut.controlMin))+onesCol*(auxOut.controlMax+auxOut.controlMin)/2;

input.phase.time    = (time_unscaled-(auxdata.tMax+auxdata.tMin)/2)./(auxdata.tMax-auxdata.tMin);
input.phase.state   = (state_unscaled-onesCol*(auxdata.stateMax+auxdata.stateMin)/2)./(onesCol*(auxdata.stateMax-auxdata.stateMin));% auxdata.Loads];% auxdata.Loadsp]
input.phase.control = (control_unscaled-onesCol*(auxdata.controlMax+auxdata.controlMin)/2)./(onesCol*(auxdata.controlMax-auxdata.controlMin));
input.auxdata = auxdata;
% 
% input = solution;
% input.auxdata = auxdata;

tracking.cont = osimOptimalTrackingContinuous(input);

input.phase.initialtime = input.phase.time(1);
input.phase.finaltime = input.phase.time(end);
input.phase.initialstate = input.phase.state(1,:);
input.phase.finalstate = input.phase.state(end,:);
input.phase.integral = trapz(input.phase.time,tracking.cont.integrand);
tracking.integral = input.phase.integral;

tracking.end = osimOptimalTrackingEndpoint(input);

tracking.time = input.phase.time;
tracking.state = input.phase.state;
tracking.control = input.phase.control;


%% Using result from prediction problem
load('solution_predictionVerification_ipopt_c1v2pt17.mat')

input.phase = solution.phase;
input.auxdata = auxdata;

prediction.cont = osimOptimalTrackingContinuous(input);

input.phase.initialtime = input.phase.time(1);
input.phase.finaltime = input.phase.time(end);
input.phase.initialstate = input.phase.state(1,:);
input.phase.finalstate = input.phase.state(end,:);
input.phase.integral = trapz(input.phase.time,prediction.cont.integrand);
prediction.integral = input.phase.integral;

prediction.end = osimOptimalTrackingEndpoint(input);

prediction.time = input.phase.time;
prediction.state = input.phase.state;
prediction.control = input.phase.control;


%% Plots and outputs
trackingGRFs = tracking.cont.GRFsSplit;
predictionGRFs = prediction.cont.GRFsSplit;

figure % 1
plot(prediction.time,predictionGRFs(:,1:3)+predictionGRFs(:,4:6),'r')
hold on
plot(tracking.time,trackingGRFs(:,1:3)+trackingGRFs(:,4:6),'b')
title('Right GRFs');

figure % 2
plot(prediction.time,predictionGRFs(:,13:15)+predictionGRFs(:,16:18),'r')
hold on
plot(tracking.time,trackingGRFs(:,13:15)+trackingGRFs(:,16:18),'b')
title('Left GRFs');

figure % 3
plot(prediction.time,predictionGRFs(:,7:9)+predictionGRFs(:,10:12),'r')
hold on
plot(tracking.time,trackingGRFs(:,7:9)+trackingGRFs(:,10:12),'b')
title('Right GRMs');

figure % 4
plot(prediction.time,predictionGRFs(:,19:21)+predictionGRFs(:,22:24),'r')
hold on
plot(tracking.time,trackingGRFs(:,19:21)+trackingGRFs(:,22:24),'b')
title('Left GRMs');

figure % 5
for i = 1:31
    if ismember(i,4:6)
%         midpoint = ((max(tracking.state(:,i)) - min(tracking.state(:,i)))/2 + min(tracking.state(:,i)));
%         minpoint = midpoint - 0.1;
%         maxpoint = midpoint + 0.1;
        subplot(6,6,i)
        plot(prediction.time, prediction.state(:,i),'r')
        hold on
    %     plot(auxdata.time/auxdata.time(end), auxdata.q(:,i),'b')
        plot(tracking.time, tracking.state(:,i),'b')
%         ylim([minpoint,maxpoint]);
        title(strrep(auxdata.qLabels(i),'_',' '));
    else
%         midpoint = (180/pi)*((max(tracking.state(:,i)) - min(tracking.state(:,i)))/2 + min(tracking.state(:,i)));
%         minpoint = midpoint - 0.75*180/pi;
%         maxpoint = midpoint + 0.75*180/pi;
        subplot(6,6,i)
        plot(prediction.time, prediction.state(:,i),'r')
        hold on
    %     plot(auxdata.time/auxdata.time(end), auxdata.q(:,i),'b')
        plot(tracking.time, tracking.state(:,i),'b')
%         ylim([minpoint,maxpoint]);
        title(strrep(auxdata.qLabels(i),'_',' '));
    end
end


trackingCost = tracking.end.objective
predictionCost = prediction.end.objective

figure;
plot(tracking.integral,'bx')
hold on
plot(prediction.integral,'r*')

trackingCostContributions = [tracking.integral(1:12)*0, ...
    tracking.integral([13,20,21:31])*1, ... %21:23,25,29
    tracking.integral(14:19)*0, ...
    (tracking.integral(32:37).*[0,0.1,0.1,0,0.1,0.1])*0, ...
    (tracking.integral(38:43).*[0.2,0.2,0,0.2,0.2,0])*0, ...
    (tracking.integral(44:55))*0, ...
    (tracking.integral(56:65))*1, ...
    (tracking.integral(66:77))*1, ...
    (tracking.integral(78:end))*0.001];
predictionCostContributions = [prediction.integral(1:12)*0, ...
    prediction.integral([13,20,21:31])*1, ... %21:23,25,29
    prediction.integral(14:19)*0, ...
    (prediction.integral(32:37).*[0,0.1,0.1,0,0.1,0.1])*0, ...
    (prediction.integral(38:43).*[0.2,0.2,0,0.2,0.2,0])*0, ...
    (prediction.integral(44:55))*0, ...
    (prediction.integral(56:65))*1, ...
    (prediction.integral(66:77))*1, ...
    (prediction.integral(78:end))*0.001];

figure;
plot(trackingCostContributions,'bx')
hold on
plot(predictionCostContributions,'r*')

figure;
for iii = 1:6
    subplot(2,3,iii)
    plot(tracking.time,tracking.cont.path(:,iii),'b');
    hold on;
    plot(prediction.time,prediction.cont.path(:,iii),'r');
end


